@@ -15,7 +15,7 @@
    limitations under the License.
 */
 
-/* $Id: mod_rivet.c 1649980 2015-01-07 00:34:23Z mxmanghi $ */
+/* $Id: mod_rivet.c 1677857 2015-05-05 16:24:36Z mxmanghi $ */
 
 /* Rivet config */
 #ifdef HAVE_CONFIG_H
@@ -101,6 +101,32 @@ static apr_status_t Rivet_ChildExit(void *data);
 
 mod_rivet_globals* rivet_module_globals = NULL;
 
+/* -- Rivet_PrintErrorMessage
+ *
+ * Utility function to print the error message stored in errorInfo
+ * with a custom header. This procedure is called to print standard
+ * errors when one of Tcl scripts fails
+ * 
+ * Arguments:
+ *
+ *   - Tcl_Interp* interp: the Tcl interpreter that was running the script
+ *                         (and therefore the built-in variable errorInfo
+ *                          keeps the message)
+ *   - const char* error: Custom error header
+ */
+
+static void 
+Rivet_PrintErrorMessage (Tcl_Interp* interp,const char* error_header)
+{
+    Tcl_Obj* errormsg  = Tcl_NewObj();
+
+    Tcl_IncrRefCount(errormsg);
+    Tcl_AppendStringsToObj(errormsg,"puts \"",error_header,"<br />\"\n",NULL);
+    Tcl_AppendStringsToObj(errormsg,"puts \"<pre>$errorInfo</pre>\"\n",NULL);
+    Tcl_EvalObjEx(interp,errormsg,0);
+    Tcl_DecrRefCount(errormsg);
+}
+
 /*
  * -- Rivet_chdir_file (const char* filename)
  * 
@@ -299,6 +325,65 @@ Rivet_InitServerVariables( Tcl_Interp *interp, apr_pool_t *pool )
     }
 }
 
+/* -- Rivet_ExecuteErrorHandler
+ *
+ * Invoking either the default error handler or the ErrorScript.
+ * In case the error handler fails a standard error message is printed
+ * (you're better off if you make your error handlers fail save)
+ *
+ * Arguments:
+ *
+ *  - Tcl_Interp* interp: the Tcl interpreter
+ *  - Tcl_Obj*    tcl_script_obj: the script that failed (to retrieve error
+ *                info from it
+ *  - request_rec* req: current request obj pointer
+ * 
+ * Returned value:
+ *
+ *  - A Tcl status
+ */
+
+static int 
+Rivet_ExecuteErrorHandler (Tcl_Interp* interp,Tcl_Obj* tcl_script_obj, rivet_interp_globals* globals)
+{
+    int                     result;
+    rivet_server_conf*      conf = Rivet_GetConf(globals->r);
+    Tcl_Obj*                errscript;
+
+    /* We extract information from the errorOutbuf variable. Notice that tcl_script_obj
+     * can either be the request processing script or conf->rivet_abort_script
+     */
+
+    Tcl_SetVar(interp, "errorOutbuf",Tcl_GetStringFromObj( tcl_script_obj, NULL ),TCL_GLOBAL_ONLY );
+
+    /* If we don't have an error script, use the default error handler. */
+    if (conf->rivet_error_script) {
+        errscript = conf->rivet_error_script;
+    } else {
+        errscript = conf->rivet_default_error_script;
+    }
+
+    Tcl_IncrRefCount(errscript);
+    result = Tcl_EvalObjEx(interp, errscript, 0);
+    if (result == TCL_ERROR) {
+        Rivet_PrintErrorMessage(interp,"<b>Rivet ErrorScript failed</b>");
+    }
+
+    /* This shouldn't make the default_error_script go away,
+     * because it gets a Tcl_IncrRefCount when it is created.
+     */
+    Tcl_DecrRefCount(errscript);
+
+    /* In case we are handling an error occurring after an abort_page call (for
+     * example because the AbortScript or AfterEveryScript themselves failed) we must
+     * reset this flag or the logging will be inihibited
+     */
+
+    globals->page_aborting = 0;
+    return result;
+}
+
+
 /* -- Rivet_ExecuteAndCheck
  * 
  * Tcl script execution central procedure. The script stored
@@ -316,7 +401,8 @@ Rivet_InitServerVariables( Tcl_Interp *interp, apr_pool_t *pool )
  *
  *   Returned value:
  *
- *      - invariably TCL_OK
+ *      - One of the Tcl defined returned value of Tcl_EvelObjExe (TCL_OK, 
+ *        TCL_ERROR, TCL_BREAK etc.)
  *
  *   Side effects:
  *
@@ -328,15 +414,18 @@ Rivet_InitServerVariables( Tcl_Interp *interp, apr_pool_t *pool )
 static int
 Rivet_ExecuteAndCheck(Tcl_Interp *interp, Tcl_Obj *tcl_script_obj, request_rec *req)
 {
-    rivet_server_conf *conf = Rivet_GetConf(req);
-    rivet_interp_globals *globals = Tcl_GetAssocData(interp, "rivet", NULL);
+    int                     result;
+    rivet_server_conf*      conf = Rivet_GetConf(req);
+    rivet_interp_globals*   globals = Tcl_GetAssocData( conf->server_interp, "rivet", NULL );
 
     Tcl_Preserve (interp);
-    if ( Tcl_EvalObjEx(interp, tcl_script_obj, 0) == TCL_ERROR ) {
-        Tcl_Obj *errscript;
-        Tcl_Obj *errorCodeListObj;
-        Tcl_Obj *errorCodeElementObj;
-        char *errorCodeSubString;
+    result = Tcl_EvalObjEx(interp, tcl_script_obj, 0);
+
+    if (result == TCL_ERROR) {
+
+        Tcl_Obj*    errorCodeListObj;
+        Tcl_Obj*    errorCodeElementObj;
+        char*       errorCodeSubString;
 
         /* There was an error, see if it's from Rivet and it was caused
          * by abort_page.
@@ -372,59 +461,44 @@ Rivet_ExecuteAndCheck(Tcl_Interp *interp, Tcl_Obj *tcl_script_obj, request_rec *
             {
                 if (conf->rivet_abort_script) 
                 {
+
+                    /* Ideally an AbortScript should be fail safe, but in case
+                     * it fails we give a chance to the subsequent ErrorScript
+                     * to catch this error.
+                     */
+
                     if (Tcl_EvalObjEx(interp,conf->rivet_abort_script,0) == TCL_ERROR)
                     {
-                        CONST84 char *errorinfo = Tcl_GetVar( interp, "errorInfo", 0 );
-                        TclWeb_PrintError("<b>Rivet AbortScript failed!</b>",1,globals->req);
-                        TclWeb_PrintError( errorinfo, 0, globals->req );
+
+                        /* This is not elegant, but we want to avoid to print
+                         * the default AbortScript error message if a custom ErrorScript 
+                         * will handle this error and in case print its own error message.
+                         */
+
+                        if (conf->rivet_error_script == NULL) 
+                        {
+                            Rivet_PrintErrorMessage(interp,"<b>Rivet AbortScript failed</b>");
+                        }
+                        Rivet_ExecuteErrorHandler(interp,conf->rivet_abort_script,globals);
+
                     }
                 }
-                goto good;
             }
         }
-
-        Tcl_SetVar( interp, "errorOutbuf",Tcl_GetStringFromObj( tcl_script_obj, NULL ),TCL_GLOBAL_ONLY );
-
-        /* If we don't have an error script, use the default error handler. */
-        if (conf->rivet_error_script ) {
-            errscript = conf->rivet_error_script;
-        } else {
-            errscript = conf->rivet_default_error_script;
-        }
-
-        Tcl_IncrRefCount(errscript);
-        if (Tcl_EvalObjEx(interp, errscript, 0) == TCL_ERROR) {
-            CONST84 char *errorinfo = Tcl_GetVar( interp, "errorInfo", 0 );
-            TclWeb_PrintError("<b>Rivet ErrorScript failed!</b>",1,globals->req);
-            TclWeb_PrintError( errorinfo, 0, globals->req );
-        }
-
-        /* This shouldn't make the default_error_script go away,
-         * because it gets a Tcl_IncrRefCount when it is created. */
-        Tcl_DecrRefCount(errscript);
-    }
-
-    /* Make sure to flush the output if buffer_add was the only output */
-good:
-    
-    if (conf->after_every_script) {
-        if (Tcl_EvalObjEx(interp,conf->after_every_script,0) == TCL_ERROR)
+        else if (!globals->page_aborting)
         {
-            CONST84 char *errorinfo = Tcl_GetVar( interp, "errorInfo", 0 );
-            TclWeb_PrintError("<b>Rivet AfterEveryScript failed!</b>",1,globals->req);
-            TclWeb_PrintError( errorinfo, 0, globals->req );
+            /* Invoke Rivet error handler */
+
+            Rivet_ExecuteErrorHandler(interp,tcl_script_obj,globals);
+
         }
     }
-
-    if (!globals->req->headers_set && (globals->req->charset != NULL)) {
-        TclWeb_SetHeaderType (apr_pstrcat(globals->req->req->pool,"text/html;",globals->req->charset,NULL),globals->req);
-    }
-
+    
     /* Tcl_Flush moved to the end of Rivet_SendContent */
 
     Tcl_Release(interp);
 
-    return TCL_OK;
+    return result;
 }
 
 /*
@@ -435,6 +509,11 @@ good:
  * if a Tcl script.
  *
  * This is a separate function so that it may be called from command 'parse'
+ * 
+ * Returned value:
+ *
+ *   - One of the Tcl defined values returned by Rivet_ExecuteAndCheck (TCL_OK, 
+ *     TCL_ERROR, TCL_BREAK etc.)
  *
  * Arguments:
  *
@@ -444,6 +523,7 @@ good:
  *   - toplevel:      integer to be interpreted as a boolean meaning the
  *                    file is pointed by the request. When 0 that's a subtemplate
  *                    to be parsed and executed from another template
+ *
  */
 
 int
@@ -453,7 +533,7 @@ Rivet_ParseExecFile(TclWebRequest *req, char *filename, int toplevel)
     int isNew = 0;
     int result = 0;
 
-    Tcl_Obj *outbuf = NULL;
+    Tcl_Obj *outbuf      = NULL;
     Tcl_HashEntry *entry = NULL;
 
     time_t ctime;
@@ -605,12 +685,23 @@ Rivet_ParseExecFile(TclWebRequest *req, char *filename, int toplevel)
     }
 
     rsc->user_scripts_updated = 0;
+    result = Rivet_ExecuteAndCheck(interp, outbuf, req->req);
+
+    if (result == TCL_ERROR)
     {
-        int res = 0;
-        res = Rivet_ExecuteAndCheck(interp, outbuf, req->req);
-        Tcl_DecrRefCount(outbuf);
-        return res;
-    }
+        rivet_interp_globals* globals = Tcl_GetAssocData( rsc->server_interp, "rivet", NULL );
+
+        /* If this is a toplevel script */
+
+        if (toplevel && globals->page_aborting) 
+        {
+            result = TCL_OK;
+        }
+    } 
+
+    Tcl_DecrRefCount(outbuf);
+
+    return result;
 }
 
 /*
@@ -629,7 +720,7 @@ Rivet_ParseExecFile(TclWebRequest *req, char *filename, int toplevel)
 int
 Rivet_ParseExecString (TclWebRequest* req, Tcl_Obj* inbuf)
 {
-    int res = 0;
+    int result;
     Tcl_Obj* outbuf = Tcl_NewObj();
     Tcl_Interp *interp = req->interp;
 
@@ -644,10 +735,10 @@ Rivet_ParseExecString (TclWebRequest* req, Tcl_Obj* inbuf)
 
     Tcl_AppendToObj(outbuf, "\n", -1);
 
-    res = Rivet_ExecuteAndCheck(interp, outbuf, req->req);
+    result = Rivet_ExecuteAndCheck(interp, outbuf, req->req);
     Tcl_DecrRefCount(outbuf);
 
-    return res;
+    return result;
 }
 
 static void
@@ -1229,7 +1320,7 @@ Rivet_ChildHandlers(server_rec *s, int init)
             Tcl_Release (rsc->server_interp);
         }
 
-        if (!init) 
+        if (!init && ((sr == s) || rsc->separate_channels)) 
         {
             Tcl_UnregisterChannel(rsc->server_interp,*(rsc->outchannel));
         }
@@ -1489,11 +1580,24 @@ Rivet_SendContent(request_rec *r)
     if (Rivet_ParseExecFile(globals->req, r->filename, 1) != TCL_OK)
     {
         ap_log_error(APLOG_MARK, APLOG_ERR, APR_EGENERAL, r->server, 
-                     MODNAME ": Error parsing exec file '%s': %s",
+                     MODNAME ": Error in Rivet_ParseExecFile exec file '%s': %s",
                      r->filename,
                      Tcl_GetVar(interp, "errorInfo", 0));
     }
 
+    /* We execute also the AfterEveryScript if one was set */
+
+    if (rsc->after_every_script)
+    {
+        int tcl_status;
+
+        tcl_status = Rivet_ExecuteAndCheck(interp, rsc->after_every_script,r);
+        if ((tcl_status == TCL_ERROR) && !globals->page_aborting)
+        {
+            Rivet_PrintErrorMessage(interp,"<b>Rivet AfterEveryScript failed</b>");
+        }
+    }
+
     if (request_cleanup == NULL) {
         request_cleanup = Tcl_NewStringObj("::Rivet::cleanup_request\n", -1);
         Tcl_IncrRefCount(request_cleanup);
@@ -1505,17 +1609,20 @@ Rivet_SendContent(request_rec *r)
                      Tcl_GetVar(interp, "errorInfo", 0));
     }
 
+    /* We finalize the request processing by printing the headers and flushing
+       the rivet channel internal buffer */
+
+    TclWeb_PrintHeaders(globals->req);
+    Tcl_Flush(*(rsc->outchannel));
+
     /* Reset globals */
-    Rivet_CleanupRequest( r );
+    Rivet_CleanupRequest(r);
 
     retval = OK;
 sendcleanup:
 
     /* Everything is done and we flush the rivet channel before resetting the status */
 
-    TclWeb_PrintHeaders(globals->req);
-    Tcl_Flush(*(rsc->outchannel));
-
     globals->req->content_sent = 0;
     globals->page_aborting = 0;
     if (globals->abort_code != NULL)
