@@ -17,7 +17,7 @@
    limitations under the License.
 */
 
-/* $Id: rivetCore.c 1650414 2015-01-08 23:03:12Z mxmanghi $ */
+/* $Id: rivetCore.c 1656602 2015-02-02 23:12:47Z mxmanghi $ */
 
 /* Rivet config */
 #ifdef HAVE_CONFIG_H
@@ -241,11 +241,7 @@ TCL_CMD_HEADER( Rivet_Parse )
         return TCL_ERROR;
     }
 
-    if (Rivet_ParseExecFile(globals->req, filename, 0) == TCL_OK) {
-        return TCL_OK;
-    } else {
-        return TCL_ERROR;
-    }
+    return Rivet_ParseExecFile(globals->req, filename, 0); 
 }
 
 /*
